<?php

namespace Drupal\mapkit_gmap\Form;

use Drupal\Core\Cache\CacheCollectorInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\Core\Utility\LinkGeneratorInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Configuration form for setting Google maps provider settings.
 */
class MapkitGmapSettings extends ConfigFormBase {

  /**
   * The library discovery service cache.
   *
   * @var \Drupal\Core\Cache\CacheCollectorInterface
   */
  protected $libraryCacheCollector;

  /**
   * Link generator service.
   *
   * @var \Drupal\Core\Utility\LinkGeneratorInterface
   */
  protected $linkGenerator;

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return parent::create($container)
      ->setLinkGenerator($container->get('link_generator'))
      ->setLibraryDiscoveryCache($container->get('library.discovery'));
  }

  /**
   * Set the link generation service.
   *
   * @param \Drupal\Core\Utility\LinkGeneratorInterface $linkGenerator
   *   The link generator service.
   *
   * @return self
   *   The instance itself for method chaining.
   */
  public function setLinkGenerator(LinkGeneratorInterface $linkGenerator): self {
    $this->linkGenerator = $linkGenerator;
    return $this;
  }

  /**
   * Set the library cache collector for the asset libraries.
   *
   * @param \Drupal\Core\Cache\CacheCollectorInterface $libraryCacheCollector
   *   The cache collector for the assets librarary discovery.
   *
   * @return self
   *   The instance itself for method chaining.
   */
  public function setLibraryDiscoveryCache(CacheCollectorInterface $libraryCacheCollector): self {
    $this->libraryCacheCollector = $libraryCacheCollector;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['mapkit_gmap.settings'];
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'mapkit_gmap_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildForm($form, $form_state);
    $config = $this->config('mapkit_gmap.settings');
    $apiUrl = Url::fromUri('https://developers.google.com/maps/documentation/javascript/get-api-key', [
      'absolute' => TRUE,
      'attributes' => ['target' => '_blank'],
    ]);

    $form['api_key'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Google Maps API key (@link)', [
        '@link' => $this->linkGenerator->generate($this->t('Get a key'), $apiUrl),
      ]),
      '#default_value' => $config->get('api_key'),
    ];

    $form['region'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Region code (optional)'),
      '#default_value' => $config->get('region'),
    ];

    $form['libraries'] = [
      '#type' => 'checkboxes',
      '#title' => $this->t('Select Google libraries to include in the Javascript APIs.'),
      '#options' => [
        'drawing' => $this->t('Drawing: polygons, circles and markers on the map'),
        'geometry' => $this->t('Geometry: calculate scalar values from geo-primatives'),
        'places' => $this->t('Places: establishments, points of interest or geographic locations'),
        'visualization' => $this->t('Visualization: heatmaps, and data representation'),
      ],
      '#default_value' => $config->get('libraries') ?: [],
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    parent::submitForm($form, $form_state);

    $values = $form_state->getValues();

    // Apply our library key values from the form.
    $this->config('mapkit_gmap.settings')->setData([
      'api_key' => $values['api_key'],
      'region' => $values['region'],
      'libraries' => array_filter($values['libraries']),
    ])->save();

    $this->libraryCacheCollector->clear();
  }

}
